#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lc.h"
#include "opcodes.h"
#include "util.h"
#include "eval.h"


int msg_level = MSG_NOTICE;


void
lowercase(char *s)
{
  int i = 0;

  while (s[i] != '\0') {
    s[i] = tolower(s[i]);
    i++;
  }
}


int parseNumber(const char *s)
{
  int x;

  // Make sure we actually have a string to parse
  if ((s == NULL) || (s[0] == '\0')) return EVAL_FAIL;

  if (isdigit(s[0])) {
    if ((s[0] == '0') && (tolower(s[1]) == 'b')) {
      // Parse as binary: 0b00110101
      MESSAGE(MSG_DEBUG, "Evaluating binary number '%s'", s);
      return strtol(&(s[2]), NULL, 2);
    }
    if (strchr(s, 'h') || strchr(s, 'H')) {
      // Parse as hex: 35h
      return strtol(s, NULL, 16);
    }
    // Otherwise, let sscanf have a try
    // This will conver decimal (123), octal (065), and hex (0x35)
    //   Negative decimals are parsed further down, since they
    //   fail the isdigit(s[0]) test above.
    sscanf(s, "%i", &x);
    return x;
  }


  if ((tolower(s[0]) == 'b') && (s[1] == '\'')) {
    // Another binary notation: b'00110101'
    return strtol(&(s[2]), NULL, 2);
  }

  // Only other possibility is a negative decimal number.
  // 's' has to be at least 2 characters long for that.
  if (s[1] == '\0') return EVAL_FAIL;
  if ((s[0] == '-') && isdigit(s[1])) {
    sscanf(s, "%i", &x);
    return x;
  }

  // If we get here, we couldn't get a valid number format.
  return EVAL_FAIL;
}


unsigned long int FNV_hash(void * dataToHash, unsigned long int length)
{
  unsigned char *p = (unsigned char *) dataToHash;
  unsigned long int h = 2166136261UL;
  unsigned long int i;

  for (i=0; i<length; i++)
      h = (h ^ p[i]) * 16777619;

  return h;
}


unsigned char Pearson_hash(void *data, int len)
{
  unsigned char *x = (unsigned char *) data;
  int i;
  unsigned char h = 0;
  static const unsigned char T[256] = {
     // 0-255 shuffled in any (random) order suffices
      98,  6, 85,150, 36, 23,112,164,135,207,169,  5, 26, 64,165,219, //  1
      61, 20, 68, 89,130, 63, 52,102, 24,229,132,245, 80,216,195,115, //  2
      90,168,156,203,177,120,  2,190,188,  7,100,185,174,243,162, 10, //  3
     237, 18,253,225,  8,208,172,244,255,126,101, 79,145,235,228,121, //  4
     123,251, 67,250,161,  0,107, 97,241,111,181, 82,249, 33, 69, 55, //  5
      59,153, 29,  9,213,167, 84, 93, 30, 46, 94, 75,151,114, 73,222, //  6
     197, 96,210, 45, 16,227,248,202, 51,152,252,125, 81,206,215,186, //  7
      39,158,178,187,131,136,  1, 49, 50, 17,141, 91, 47,129, 60, 99, //  8
     154, 35, 86,171,105, 34, 38,200,147, 58, 77,118,173,246, 76,254, //  9
     133,232,196,144,198,124, 53,  4,108, 74,223,234,134,230,157,139, // 10
     189,205,199,128,176, 19,211,236,127,192,231, 70,233, 88,146, 44, // 11
     183,201, 22, 83, 13,214,116,109,159, 32, 95,226,140,220, 57, 12, // 12
     221, 31,209,182,143, 92,149,184,148, 62,113, 65, 37, 27,106,166, // 13
       3, 14,204, 72, 21, 41, 56, 66, 28,193, 40,217, 25, 54,179,117, // 14
     238, 87,240,155,180,170,242,212,191,163, 78,218,137,194,175,110, // 15
      43,119,224, 71,122,142, 42,160,104, 48,247,103, 15, 11,138,239  // 16
  };
  for (i=0; i<len; i++) {
    h = T[h ^ x[i]];
  }
  return h;
}


void message(int level, const char *fmt, ...)
{
  va_list args;

  if (level > msg_level) return;

  va_start(args, fmt);
  fprintf(stderr, "%04x ", lcGet());
  switch (level) {
    case MSG_CRITICAL: fprintf(stderr, "CRITICAL ERROR: "); break;
    case MSG_ERROR:    fprintf(stderr, "ERROR: ");   break;
    case MSG_WARNING:  fprintf(stderr, "WARNING: "); break;
    case MSG_NOTICE:   fprintf(stderr, "NOTICE: ");  break;
    case MSG_INFO:     fprintf(stderr, "INFO: ");    break;
    case MSG_VERBOSE:  fprintf(stderr, "INFO: ");    break;
    case MSG_DEBUG:    fprintf(stderr, "DEBUG: ");   break;
    case MSG_VDEBUG:   fprintf(stderr, "DEBUG: ");   break;
  }
  vfprintf(stderr, fmt, args);
  fprintf(stderr, "\n");
  va_end(args);
}


unsigned int shorten2s(const int x, const int bits)
{
  int sign = (x < 0) ? 1 : 0;
  unsigned int r;

  if (x >= 0) return x & ((1 << (bits-1)) - 1);

  r = abs(x);
  r = ~r;
  r++;
  r = r & ((1 << (bits-1)) - 1);
  r = r | (sign << (bits-1));
  return r;
}


int extendsign(const unsigned int x, const int bits)
{
  int i;
  int bm;	// bitmask
  int r;
  int sign = (x & (1 << (bits-1))) ? 1 : 0;
  // int r = (x & ((1 << (bits-1)) - 1));
  r = x;
  for (i=bits; i<=(sizeof(int)*8); i++) {
    bm = (1 << (i-1));
    r = (r & ~bm);	// Clear bit
    if (!sign) bm = 0;
    r = (r | bm);
  }

  return r;
}

